# -*- coding: utf-8 -*-

# Copyright 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide an interface for essentiality experiments."""

from __future__ import absolute_import

import logging
from typing import Optional

import pandera as pa
from cobra.flux_analysis import single_gene_deletion
from pandera.typing import Series

from memote.experimental.experiment import Experiment


__all__ = ("EssentialityExperiment",)

LOGGER = logging.getLogger(__name__)


class EssentialityExperimentModel(pa.DataFrameModel):
    gene: Series[str] = pa.Field(
        title="Gene Identifier",
        description="The gene identifier must correspond to the metabolic model "
        "identifiers.",
        unique=True,
    )
    essential: Series[bool] = pa.Field(
        title="Gene Essentiality",
        description="Whether a gene is (conditionally) essential.",
    )
    comment: Optional[Series[str]] = pa.Field(
        nullable=True,
        title="Comment",
        description="Optional comment which is not processed further.",
    )

    class Config:
        coerce = True
        strict = "filter"


class EssentialityExperiment(Experiment):
    """Represent an essentiality experiment."""

    def __init__(self, **kwargs):
        """
        Initialize an essentiality experiment.

        Parameters
        ----------
        kwargs

        """
        super(EssentialityExperiment, self).__init__(**kwargs)

    def validate(self, model, checks=None):
        """Use a defined schema to validate the essentiality table format."""
        EssentialityExperimentModel.validate(self.data, lazy=True)
        assert self.data["gene"].isin({g.id for g in model.genes}).all()

    def evaluate(self, model):
        """Use the defined parameters to predict single gene essentiality."""
        with model:
            if self.medium is not None:
                self.medium.apply(model)
            if self.objective is not None:
                model.objective = self.objective
            model.add_cons_vars(self.constraints)
            essen = single_gene_deletion(
                model, gene_list=self.data["gene"], processes=1
            )
        essen["gene"] = [list(g)[0] for g in essen["ids"]]
        essen["essential"] = (essen["growth"] < self.minimal_growth_rate) | essen[
            "growth"
        ].isna()
        return essen



# -*- coding: utf-8 -*-

# Copyright 2017 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide a YAML parser for experiment configuration files."""

from __future__ import absolute_import

import json
import logging


try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

from io import open
from math import isnan
from os.path import dirname, isabs, join

from jsonschema import Draft4Validator
from ruamel.yaml import YAML
from six import iteritems

from memote.experimental.essentiality import EssentialityExperiment
from memote.experimental.growth import GrowthExperiment
from memote.experimental.medium import Medium


__all__ = ("ExperimentConfiguration",)

LOGGER = logging.getLogger(__name__)
yaml = YAML(typ="safe")


class ExperimentConfiguration(object):
    """Represent an experimental configuration."""

    with files("memote.experimental.schemata").joinpath("configuration.json").open(
        mode="r", encoding="utf-8"
    ) as file_handle:
        SCHEMA = json.load(file_handle)

    def __init__(self, filename, **kwargs):
        """
        Load and validate an experiment configuration file.

        Parameters
        ----------
        filename : str or pathlib.Path
            The location of the configuration file.

        """
        super(ExperimentConfiguration, self).__init__(**kwargs)
        with open(filename) as file_h:
            self.config = yaml.load(file_h)
        self._base = dirname(filename)
        self.media = dict()
        self.essentiality = dict()
        self.growth = dict()

    def load(self, model):
        """
        Load all information from an experimental configuration file.

        Parameters
        ----------
        model : cobra.Model
            The metabolic model under investigation.

        """
        self.load_medium(model)
        self.load_essentiality(model)
        self.load_growth(model)
        # self.load_experiment(config.config.get("growth"), model)
        return self

    def validate(self):
        """Validate the configuration file."""
        validator = Draft4Validator(self.SCHEMA)
        if not validator.is_valid(self.config):
            for err in validator.iter_errors(self.config):
                LOGGER.error(str(err.message))
            validator.validate(self.config)

    def load_medium(self, model):
        """Load and validate all media."""
        media = self.config.get("medium")
        if media is None:
            return
        definitions = media.get("definitions")
        if definitions is None or len(definitions) == 0:
            return
        path = self.get_path(media, join("data", "experimental", "media"))
        for medium_id, medium in iteritems(definitions):
            if medium is None:
                medium = dict()
            filename = medium.get("filename")
            if filename is None:
                filename = join(path, "{}.csv".format(medium_id))
            elif not isabs(filename):
                filename = join(path, filename)
            tmp = Medium(identifier=medium_id, obj=medium, filename=filename)
            tmp.load()
            tmp.validate(model)
            self.media[medium_id] = tmp

    def load_essentiality(self, model):
        """Load and validate all data files."""
        data = self.config.get("essentiality")
        if data is None:
            return
        experiments = data.get("experiments")
        if experiments is None or len(experiments) == 0:
            return
        path = self.get_path(data, join("data", "experimental", "essentiality"))
        minimal_growth_rate = self.get_minimal_growth_rate(model)
        for exp_id, exp in iteritems(experiments):
            if exp is None:
                exp = dict()
            filename = exp.get("filename")
            if filename is None:
                filename = join(path, "{}.csv".format(exp_id))
            elif not isabs(filename):
                filename = join(path, filename)
            experiment = EssentialityExperiment(
                identifier=exp_id,
                obj=exp,
                filename=filename,
                minimal_growth_rate=minimal_growth_rate,
            )
            if experiment.medium is not None:
                assert (
                    experiment.medium in self.media
                ), "Experiment '{}' has an undefined medium '{}'.".format(
                    exp_id, experiment.medium
                )
                experiment.medium = self.media[experiment.medium]
            experiment.load()
            experiment.validate(model)
            self.essentiality[exp_id] = experiment

    def load_growth(self, model):
        """Load and validate all data files."""
        data = self.config.get("growth")
        if data is None:
            return
        experiments = data.get("experiments")
        if experiments is None or len(experiments) == 0:
            return
        path = self.get_path(data, join("data", "experimental", "growth"))
        minimal_growth_rate = self.get_minimal_growth_rate(model)
        for exp_id, exp in iteritems(experiments):
            if exp is None:
                exp = dict()
            filename = exp.get("filename")
            if filename is None:
                filename = join(path, "{}.csv".format(exp_id))
            elif not isabs(filename):
                filename = join(path, filename)
            growth = GrowthExperiment(
                identifier=exp_id,
                obj=exp,
                filename=filename,
                minimal_growth_rate=minimal_growth_rate,
            )
            if growth.medium is not None:
                assert (
                    growth.medium in self.media
                ), "Growth-experiment '{}' has an undefined medium '{}'.".format(
                    exp_id, growth.medium
                )
                growth.medium = self.media[growth.medium]
            growth.load()
            growth.validate(model)
            self.growth[exp_id] = growth

    def get_path(self, obj, default):
        """Return a relative or absolute path to experimental data."""
        path = obj.get("path")
        if path is None:
            path = join(self._base, default)
        if not isabs(path):
            path = join(self._base, path)
        return path

    def get_minimal_growth_rate(self, model, threshold=0.1):
        """Calculate min growth default value or return input value.

        This value is used to determine if a model is capable of growth under
        certain experimental conditions.

        Parameters
        ----------
        model : cobra.Model
        threshold : float, optional
            If no input is provided by the user the default value is set to a
            coefficient `threshold` times the growth under default constraints
            (default: 0.1).

        """
        minimal_growth_rate = self.config.get("minimal_growth_rate")
        if minimal_growth_rate is None:
            minimal_growth_rate = model.slim_optimize() * threshold
            if isnan(minimal_growth_rate):
                LOGGER.error(
                    "Threshold set to {} due to infeasible "
                    "solution (NaN produced) with default "
                    "constraints.".format(model.tolerance)
                )
                minimal_growth_rate = model.tolerance
        return minimal_growth_rate



# -*- coding: utf-8 -*-

# Copyright 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide a class for medium definitions."""

from __future__ import absolute_import

import logging
from typing import Optional

import pandera as pa
from pandera.typing import Series

from memote.experimental.experimental_base import ExperimentalBase


__all__ = ("Medium",)

LOGGER = logging.getLogger(__name__)


class MediumModel(pa.DataFrameModel):
    exchange: Series[str] = pa.Field(
        description="The exchange reaction identifiers must correspond to the "
        "metabolic model identifiers.",
        title="Exchange Reaction Identifier",
        unique=True,
    )
    uptake: Series[float] = pa.Field(
        ge=0.0,
        le=1000.0,
        title="Uptake Rate",
        description="The uptake rate for the exchange reaction. For models following "
        "common practices this modifies the lower bound.",
    )
    comment: Optional[Series[str]] = pa.Field(
        nullable=True,
        title="Comment",
        description="Optional comment which is not processed further.",
    )

    class Config:
        coerce = True
        strict = "filter"


class Medium(ExperimentalBase):
    """Represent a specific medium condition."""

    def __init__(self, **kwargs):
        """
        Initialize a medium.

        Parameters
        ----------
        kwargs

        """
        super(Medium, self).__init__(**kwargs)

    def validate(self, model):
        """Use a defined schema to validate the medium table format."""
        MediumModel.validate(self.data, lazy=True)
        assert self.data["exchange"].isin({r.id for r in model.reactions}).all()

    def apply(self, model):
        """Set the defined medium on the given model."""
        model.medium = {
            row.exchange: row.uptake for row in self.data.itertuples(index=False)
        }



# -*- coding: utf-8 -*-

# Copyright 2017 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Callbacks for command line parameters."""

from __future__ import absolute_import

import logging
import shlex
import sys
from subprocess import CalledProcessError, check_output

import click
import git

from memote.experimental import ExperimentConfiguration


LOGGER = logging.getLogger(__name__)
DEFAULT_SOLVERS = ("glpk", "cplex", "gurobi", "glpk_exact", "hybrid")


def validate_collect(context, param, value):
    """Handle the report collection flag."""
    if value is not None:
        return value
    else:
        return True


def validate_experimental(context, param, value):
    """Load and validate an experimental data configuration."""
    if value is None:
        return
    config = ExperimentConfiguration(value)
    config.validate()
    return config


def validate_pytest_args(context, param, value):
    """Handle additional arguments to pytest."""
    if value is not None:
        return shlex.split(value)
    else:
        return list()


def validate_repository(context, param, value):
    """Load repository slug from different locations."""
    if value is not None:
        return value
    else:
        raise click.BadParameter("No GitHub repository slug provided or configured.")


def validate_token(context, param, value):
    """Verify that a token was provided, pointing to documentation."""
    if value is not None:
        return value
    else:
        raise click.BadParameter(
            "No GitHub token provided. See https://memote.readthedocs.io/en/"
            "latest/getting_started.html#ci-tested-online-and-public-workflow"
            " for instructions on how to set it up."
        )


def probe_git():
    """Return a git repository instance if it exists."""
    try:
        repo = git.Repo()
    except git.InvalidGitRepositoryError:
        LOGGER.warning(
            "We highly recommend keeping your model in a git repository."
            " It allows you to track changes and to easily collaborate with"
            " others via online platforms such as https://github.com.\n"
        )
        return
    if repo.is_dirty():
        LOGGER.critical(
            "Please git commit or git stash all changes before running"
            " the memote suite."
        )
        sys.exit(1)
    return repo


def abort_if_false(ctx, param, value):
    """Require confirmation."""
    if not value:
        ctx.abort()


def git_installed():
    """Interrupt execution of memote if `git` has not been installed."""
    LOGGER.info("Checking `git` installation.")
    try:
        check_output(["git", "--version"])
    except CalledProcessError as e:
        LOGGER.critical(
            "The execution of memote was interrupted since no installation of "
            "`git` could be detected. Please install git to use "
            "this functionality: "
            "https://git-scm.com/book/en/v2/Getting-Started-Installing-Git"
        )
        LOGGER.debug("Underlying error:", exc_info=e)
        sys.exit(1)



# -*- coding: utf-8 -*-

# Copyright 2017 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import git
import pytest

import memote.utils as utils


def one():
    """One line."""
    pass


def two():
    """Two lines.
    Why?
    """
    pass


def three():
    """
    Three lines.
    Why?
    """
    pass


def four():
    """
    Fourth summary.

    A proper description.

    Returns
    -------
    None

    """
    pass


@pytest.mark.parametrize(
    "func, func_name", [(one, "one"), (two, "two"), (three, "three"), (four, "four")]
)
def test_register_with(func, func_name):
    registry = dict()
    utils.register_with(registry)(func)
    assert func_name in registry
    assert registry[func_name] is func


@pytest.mark.parametrize(
    "notes, func, summary",
    [
        (dict(title="One", data=4, format_type="raw"), one, """One line."""),
        (
            dict(title="Two", data="some text", format_type="raw"),
            two,
            """Two lines.
    Why?
    """,
        ),
        (
            dict(title="Three", data=[2, 4, 51, 63], format_type="count"),
            three,
            """
    Three lines.
    Why?
    """,
        ),
        (
            dict(title="Four", data=None, format_type="raw"),
            four,
            """
    Fourth summary.

    A proper description.

    Returns
    -------
    None

    """,
        ),
        pytest.param(
            dict(title="Some Title", format_type="wrong_type"),
            one,
            "",
            marks=pytest.mark.raises(exception=ValueError),
        ),
    ],
)
def test_annotate(notes, func, summary):
    res = utils.annotate(**notes)(func)
    assert res.annotation["title"] == notes["title"]
    assert res.annotation["summary"] == utils.extended_summary(func)
    assert res.annotation["data"] is notes["data"]
    assert res.annotation["message"] is None
    assert res.annotation["format_type"] == notes["format_type"]
    assert res.annotation["metric"] == 1.0


def test_show_versions(capsys):
    utils.show_versions()
    captured = capsys.readouterr()
    lines = captured.out.split("\n")
    assert lines[1].startswith("Package Information")
    assert lines[2].startswith("------------------")
    assert lines[3].startswith("memote")

    assert lines[5].startswith("Dependency Information")
    assert lines[6].startswith("------------------")


@pytest.fixture(scope="function")
def mock_repo(tmpdir_factory):
    """Mock repo with history: add, modify, do nothing, delete file."""
    path = str(tmpdir_factory.mktemp("memote-mock-repo"))
    repo = git.Repo.init(path)
    with repo.config_writer() as writer:
        writer.set_value("user", "name", "memote-bot").release()
        writer.set_value("user", "email", "bot@memote.io").release()
    relname = "test_file.txt"
    absname = os.path.join(path, relname)

    for message in "Add file.", "Modify file.":
        with open(absname, "w") as f:
            f.write(message)
        repo.index.add([relname])
        repo.git.commit("--message", message)

    with open(os.path.join(path, "unrelated_file.txt"), "w") as f:
        f.write("Unrelated file.")
    repo.index.add(["unrelated_file.txt"])
    repo.git.commit("--message", "A commit that does not touch {}".format(relname))

    repo.index.remove([relname], working_tree=True)
    repo.git.commit("--message", "Delete file.")

    return relname, repo


def test_is_modified_deleted(mock_repo):
    """Don't report file deletion as modification."""
    relname, repo = mock_repo
    # File history (newest first): deleted, unchanged, modified, created
    # Don't report file deletion as "modification"
    # for purposes of running memote tests on the file.
    want = False, False, True, True
    # Convert to tuple so we can compare want == got,
    # which pytest will introspect helpfully if the assertion fails.
    got = tuple(utils.is_modified(relname, commit) for commit in repo.iter_commits())
    assert want == got



# -*- coding: utf-8 -*-

# Copyright 2017 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Programmatically interact with the memote test suite."""

from __future__ import absolute_import

import logging
import os
from io import open

import pytest
from jinja2 import Environment, PackageLoader, select_autoescape

from memote.suite import TEST_DIRECTORY
from memote.suite.collect import ResultCollectionPlugin
from memote.suite.reporting import (
    DiffReport,
    HistoryReport,
    ReportConfiguration,
    SnapshotReport,
)
from memote.support import validation as val


__all__ = (
    "validate_model",
    "test_model",
    "snapshot_report",
    "diff_report",
    "history_report",
)

LOGGER = logging.getLogger(__name__)


def validate_model(path):
    """
    Validate a model structurally and optionally store results as JSON.

    Parameters
    ----------
    path :
        Path to model file.

    Returns
    -------
    tuple
        cobra.Model
            The metabolic model under investigation.
        tuple
            A tuple reporting on the SBML level, version, and FBC package
            version used (if any) in the SBML document.
        dict
            A simple dictionary containing a list of errors and warnings.

    """
    notifications = {"warnings": [], "errors": []}
    model, sbml_ver = val.load_cobra_model(path, notifications)
    return model, sbml_ver, notifications


def test_model(
    model,
    sbml_version=None,
    results=False,
    pytest_args=None,
    exclusive=None,
    skip=None,
    experimental=None,
    solver_timeout=10,
):
    """
    Test a model and optionally store results as JSON.

    Parameters
    ----------
    model : cobra.Model
        The metabolic model under investigation.
    sbml_version: tuple, optional
        A tuple reporting on the level, version, and FBC use of the SBML file.
    results : bool, optional
        Whether to return the results in addition to the return code.
    pytest_args : list, optional
        Additional arguments for the pytest suite.
    exclusive : iterable, optional
        Names of test cases or modules to run and exclude all others. Takes
        precedence over ``skip``.
    skip : iterable, optional
        Names of test cases or modules to skip.
    solver_timeout: int, optional
        Timeout in seconds to set on the mathematical optimization solver (default 10).

    Returns
    -------
    int
        The return code of the pytest suite.
    memote.Result, optional
        A nested dictionary structure that contains the complete test results.

    """
    if pytest_args is None:
        pytest_args = ["--tb", "short", TEST_DIRECTORY]
    elif not any(a.startswith("--tb") for a in pytest_args):
        pytest_args.extend(["--tb", "short"])
    if TEST_DIRECTORY not in pytest_args:
        pytest_args.append(TEST_DIRECTORY)
    if "-s" not in pytest_args and LOGGER.getEffectiveLevel() <= logging.DEBUG:
        # Disable pytest capturing so that the solver log output can be seen
        # immediately.
        pytest_args.insert(0, "-s")
    model.solver.configuration.timeout = solver_timeout
    # Load the experimental configuration using model information.
    if experimental is not None:
        experimental.load(model)
    plugin = ResultCollectionPlugin(
        model,
        sbml_version=sbml_version,
        exclusive=exclusive,
        skip=skip,
        experimental_config=experimental,
    )
    code = pytest.main(pytest_args, plugins=[plugin])
    if results:
        return code, plugin.results
    else:
        return code


def snapshot_report(result, config=None, html=True):
    """
    Generate a snapshot report from a result set and configuration.

    Parameters
    ----------
    result : memote.MemoteResult
        Nested dictionary structure as returned from the test suite.
    config : dict, optional
        The final test report configuration (default None).
    html : bool, optional
        Whether to render the report as full HTML or JSON (default True).

    """
    if config is None:
        config = ReportConfiguration.load()
    report = SnapshotReport(result=result, configuration=config)
    if html:
        return report.render_html()
    else:
        return report.render_json()


def history_report(history, config=None, html=True):
    """
    Test a model and save a history report.

    Parameters
    ----------
    history : memote.HistoryManager
        The manager grants access to previous results.
    config : dict, optional
        The final test report configuration.
    html : bool, optional
        Whether to render the report as full HTML or JSON (default True).

    """
    if config is None:
        config = ReportConfiguration.load()
    report = HistoryReport(history=history, configuration=config)
    if html:
        return report.render_html()
    else:
        return report.render_json()


def diff_report(diff_results, config=None, html=True):
    """
    Generate a diff report from a result set and configuration.

    Parameters
    ----------
    diff_results : iterable of memote.MemoteResult
        Nested dictionary structure as returned from the test suite.
    config : dict, optional
        The final test report configuration (default None).
    html : bool, optional
        Whether to render the report as full HTML or JSON (default True).

    """
    if config is None:
        config = ReportConfiguration.load()
    report = DiffReport(diff_results=diff_results, configuration=config)
    if html:
        return report.render_html()
    else:
        return report.render_json()


def validation_report(path, notifications, filename):
    """
    Generate a validation report from a notification object.

    Parameters
    ----------
    path : string
        Path to model file.
    notifications : dict
        A simple dictionary structure containing a list of errors and warnings.

    """
    env = Environment(
        loader=PackageLoader("memote.suite", "templates"),
        autoescape=select_autoescape(["html", "xml"]),
    )
    template = env.get_template("validation_template.html")
    model = os.path.basename(path)
    with open(filename, "w") as file_h:
        file_h.write(template.render(model=model, notifications=notifications))
